@@ -1,7 +1,7 @@
 /*
  * Copyright (c) 1996, 1997, 1998, 1999
  *             Shigio Yamaguchi. All rights reserved.
- * Copyright (c) 1999, 2000, 2001
+ * Copyright (c) 1999, 2000, 2001, 2002
  *             Tama Communications Corporation. All rights reserved.
  *
  * This file is part of GNU GLOBAL.
@@ -36,9 +36,8 @@
 #define GTAGSRC         ".globalrc"
 #define DEFAULTLABEL    "default"
 #define DEFAULTSUFFIXES "c,h,y,c++,cc,cpp,cxx,hxx,C,H,s,S,java"
-#define DEFAULTSKIP     "y.tab.c,y.tab.h,SCCS/,RCS/,CVS/,.deps/"
+#define DEFAULTSKIP     "GPATH,GTAGS,GRTAGS,GSYMS,HTML/,tags,TAGS,ID,y.tab.c,y.tab.h,SCCS/,RCS/,CVS/,CVSROOT/,.deps/"
 
-char	*configpath(void);
 void	openconf(void);
 int	getconfn(const char *, int *);
 int	getconfs(const char *, STRBUF *);
