@@ -1,7 +1,7 @@
 /*
  * Copyright (c) 1996, 1997, 1998, 1999
  *             Shigio Yamaguchi. All rights reserved.
- * Copyright (c) 1999, 2000
+ * Copyright (c) 1999, 2000, 2001, 2002
  *             Tama Communications Corporation. All rights reserved.
  *
  * This file is part of GNU GLOBAL.
@@ -55,6 +55,7 @@
 #include "conf.h"
 #include "die.h"
 #include "find.h"
+#include "is_unixy.h"
 #include "locatestring.h"
 #include "makepath.h"
 #include "strbuf.h"
@@ -70,12 +71,17 @@
  *
  */
 static regex_t	skip_area;
-static regex_t	*skip = &skip_area;
+static regex_t	*skip;			/* regex for skipping units */
+static regex_t	suff_area;
+static regex_t	*suff = &suff_area;	/* regex for suffixes */
+static STRBUF *list;
+static int list_count;
+static char **listarray;		/* list for skipping full path */
 static int	opened;
 static int	retval;
 
 static void	trim(char *);
-
+extern int	debug;
 /*
  * trim: remove blanks and '\'.
  */
@@ -94,6 +100,211 @@ char	*s;
 	}
 	*p = 0;
 }
+/*
+ * prepare_source: preparing regular expression.
+ *
+ *	i)	flags	flags for regcomp.
+ *	go)	suff	regular expression for source files.
+ */
+static void
+prepare_source()
+{
+	STRBUF	*sb = strbuf_open(0);
+	char	*sufflist = NULL;
+	int	flags = REG_EXTENDED;
+
+	/*
+	 * load icase_path option.
+	 */
+	if (getconfb("icase_path"))
+		flags |= REG_ICASE;
+#if defined(_WIN32) || defined(__DJGPP__)
+	flags |= REG_ICASE;
+#endif
+	strbuf_reset(sb);
+	if (!getconfs("suffixes", sb))
+		die("cannot get suffixes data.");
+	sufflist = strdup(strbuf_value(sb));
+	if (!sufflist)
+		die("short of memory.");
+	trim(sufflist);
+	{
+		char    *suffp;
+
+		strbuf_reset(sb);
+		strbuf_puts(sb, "\\.(");       /* ) */
+		for (suffp = sufflist; suffp; ) {
+			char    *p;
+
+			for (p = suffp; *p && *p != ','; p++) {
+				strbuf_putc(sb, '\\');
+				strbuf_putc(sb, *p);
+			}
+			if (!*p)
+				break;
+			assert(*p == ',');
+			strbuf_putc(sb, '|');
+			suffp = ++p;
+		}
+		strbuf_puts(sb, ")$");
+		/*
+		 * compile regular expression.
+		 */
+		retval = regcomp(suff, strbuf_value(sb), flags);
+		if (debug)
+			fprintf(stderr, "find regex: %s\n", strbuf_value(sb));
+		if (retval != 0)
+			die("cannot compile regular expression.");
+	}
+	strbuf_close(sb);
+	if (sufflist)
+		free(sufflist);
+}
+/*
+ * prepare_skip: prepare skipping files.
+ *
+ *	go)	skip	regular expression for skip files.
+ *	go)	listarry[] skip list.
+ *	go)	list_count count of skip list.
+ */
+static void
+prepare_skip()
+{
+	char *skiplist;
+	STRBUF *reg = strbuf_open(0);
+	int reg_count = 0;
+	char *p, *q;
+	int     flags = REG_EXTENDED|REG_NEWLINE;
+
+	/*
+	 * load icase_path option.
+	 */
+	if (getconfb("icase_path"))
+		flags |= REG_ICASE;
+#if defined(_WIN32) || defined(__DJGPP__)
+	flags |= REG_ICASE;
+#endif
+	/*
+	 * initinalize common data.
+	 */
+	if (!list)
+		list = strbuf_open(0);
+	else
+		strbuf_reset(list);
+	list_count = 0;
+	if (listarray)
+		(void)free(listarray);
+	listarray = (char **)0;
+	/*
+	 * load skip data.
+	 */
+	if (!getconfs("skip", reg))
+		return;
+	skiplist = strdup(strbuf_value(reg));
+	if (!skiplist)
+		die("short of memory.");
+	trim(skiplist);
+	strbuf_reset(reg);
+	/*
+	 * construct regular expression.
+	 */
+	strbuf_putc(reg, '(');	/* ) */
+	for (p = skiplist; p; ) {
+		char    *skipf = p;
+		if ((p = locatestring(p, ",", MATCH_FIRST)) != NULL)
+			*p++ = 0;
+		if (*skipf == '/') {
+			list_count++;
+			strbuf_puts0(list, skipf);
+		} else {
+			reg_count++;
+			strbuf_putc(reg, '/');
+			for (q = skipf; *q; q++) {
+				if (*q == '.')
+					strbuf_putc(reg, '\\');
+				strbuf_putc(reg, *q);
+			}
+			if (*(q - 1) != '/')
+				strbuf_putc(reg, '$');
+			if (p)
+				strbuf_putc(reg, '|');
+		}
+	}
+	strbuf_unputc(reg, '|');
+	strbuf_putc(reg, ')');
+	if (reg_count > 0) {
+		/*
+		 * compile regular expression.
+		 */
+		skip = &skip_area;
+		retval = regcomp(skip, strbuf_value(reg), flags);
+		if (debug)
+			fprintf(stderr, "skip regex: %s\n", strbuf_value(reg));
+		if (retval != 0)
+			die("cannot compile regular expression.");
+	} else {
+		skip = (regex_t *)0;
+	}
+	if (list_count > 0) {
+		int i;
+		listarray = (char **)malloc(sizeof(char *) * list_count);
+		p = strbuf_value(list);
+		if (debug)
+			fprintf(stderr, "skip list: ");
+		for (i = 0; i < list_count; i++) {
+			if (debug) {
+				fprintf(stderr, "%s", p);
+				if (i + 1 < list_count)
+					fputc(',', stderr);
+			}
+			listarray[i] = p;
+			p += strlen(p) + 1;
+		}
+		if (debug)
+			fputc('\n', stderr);
+	}
+	strbuf_close(reg);
+	free(skiplist);
+}
+/*
+ * skipthisfile: check whether or not we accept this file.
+ *
+ *	i)	path	path name (must start with ./)
+ *	r)		1: skip, 0: dont skip
+ */
+int
+skipthisfile(path)
+char *path;
+{
+	char *first, *last;
+	int i;
+
+	/*
+	 * unit check.
+	 */
+	if (skip && regexec(skip, path, 0, 0, 0) == 0)
+		return 1;
+	/*
+	 * list check.
+	 */
+	if (list_count == 0)
+		return 0;
+	for (i = 0; i < list_count; i++) {
+		first = listarray[i];
+		last = first + strlen(first);
+		/*
+		 * the path must start with "./".
+		 */
+		if (*(last - 1) == '/') {	/* it's a directory */
+			if (!strncmp(path + 1, first, last - first))
+				return 1;
+		} else {
+			if (!strcmp(path + 1, first))
+				return 1;
+		}
+	}
+	return 0;
+}
 #ifndef HAVE_FIND
 /*----------------------------------------------------------------------*/
 /* dirent version findxxx()						*/
@@ -105,11 +316,19 @@ static  struct {
 	char    *dirp, *start, *end, *p;
 } stack[STACKSIZE], *topp, *curp;		/* stack */
 
-static regex_t	suff_area;
-static regex_t	*suff = &suff_area;
-
 static int getdirs(char *, STRBUF *);
 
+/*
+ * getdirs: get directory list
+ *
+ *	i)	dir	directory
+ *	o)	sb	string buffer
+ *	r)		-1: error, 0: normal
+ *
+ * format of directory list:
+ * |ddir1\0ffile1\0llink\0|
+ * means directory 'dir1', file 'file1' and symbolic link 'link'.
+ */
 static int
 getdirs(dir, sb)
 char    *dir;
@@ -175,25 +394,15 @@ STRBUF  *sb;
 	(void)closedir(dirp);
 	return 0;
 }
+/*
+ * find_open: start iterator without GPATH.
+ */
 void
-ffindopen()
+find_open()
 {
-	STRBUF	*sb = strbuf_open(0);
-	char	*sufflist = NULL;
-	char	*skiplist = NULL;
-	int	flags = REG_EXTENDED;
-
 	assert(opened == 0);
 	opened = 1;
 
-	/*
-	 * load icase_path option.
-	 */
-	if (getconfb("icase_path"))
-		flags |= REG_ICASE;
-#ifdef _WIN32
-	flags |= REG_ICASE;
-#endif /* _WIN32 */
 	/*
 	 * setup stack.
 	 */
@@ -209,89 +418,18 @@ ffindopen()
 	curp->end   = curp->start + strbuf_getlen(curp->sb);
 
 	/*
-	 * preparing regular expression.
+	 * prepare regular expressions.
 	 */
-	strbuf_reset(sb);
-	if (!getconfs("suffixes", sb))
-		die("cannot get suffixes data.");
-	sufflist = strdup(strbuf_value(sb));
-	if (!sufflist)
-		die("short of memory.");
-	trim(sufflist);
-	strbuf_reset(sb);
-	if (getconfs("skip", sb)) {
-		skiplist = strdup(strbuf_value(sb));
-		if (!skiplist)
-			die("short of memory.");
-		trim(skiplist);
-	}
-	{
-		char    *suffp;
-
-		strbuf_reset(sb);
-		strbuf_puts(sb, "\\.(");       /* ) */
-		for (suffp = sufflist; suffp; ) {
-			char    *p;
-
-			for (p = suffp; *p && *p != ','; p++) {
-				strbuf_putc(sb, '\\');
-				strbuf_putc(sb, *p);
-			}
-			if (!*p)
-				break;
-			assert(*p == ',');
-			strbuf_putc(sb, '|');
-			suffp = ++p;
-		}
-		strbuf_puts(sb, ")$");
-		/*
-		 * compile regular expression.
-		 */
-		retval = regcomp(suff, strbuf_value(sb), flags);
-		if (retval != 0)
-			die("cannot compile regular expression.");
-	}
-	if (skiplist) {
-		char    *p, *q;
-		/*
-		 * construct regular expression.
-		 */
-		strbuf_reset(sb);
-		strbuf_putc(sb, '(');	/* ) */
-		for (p = skiplist; p; ) {
-			char    *skipf = p;
-			if ((p = locatestring(p, ",", MATCH_FIRST)) != NULL)
-				*p++ = 0;
-			strbuf_putc(sb, '/');
-			for (q = skipf; *q; q++) {
-				if (*q == '.')
-					strbuf_putc(sb, '\\');
-				strbuf_putc(sb, *q);
-			}
-			if (*(q - 1) != '/')
-				strbuf_putc(sb, '$');
-			if (p)
-				strbuf_putc(sb, '|');
-		}
-		strbuf_putc(sb, ')');
-		/*
-		 * compile regular expression.
-		 */
-		retval = regcomp(skip, strbuf_value(sb), flags);
-		if (retval != 0)
-			die("cannot compile regular expression.");
-	} else {
-		skip = (regex_t *)0;
-	}
-	strbuf_close(sb);
-	if (sufflist)
-		free(sufflist);
-	if (skiplist)
-		free(skiplist);
+	prepare_source();
+	prepare_skip();
 }
+/*
+ * find_read: read path without GPATH.
+ *
+ *	r)		path
+ */
 char    *
-ffindread(length)
-int	*length;
+find_read(void)
 {
 	static	char val[MAXPATHLEN+1];
 
@@ -303,13 +441,16 @@ int	*length;
 			curp->p += strlen(curp->p) + 1;
 			if (type == 'f' || type == 'l') {
 				char	*path = makepath(dir, unit, NULL);
-				if (regexec(suff, path, 0, 0, 0) != 0)
+				if (skipthisfile(path))
 					continue;
-				if (skip && regexec(skip, path, 0, 0, 0) == 0)
-					continue;
-				strcpy(val, path);
-				if (length)
-					*length = strlen(val);
+				if (regexec(suff, path, 0, 0, 0) == 0) {
+					/* source file */
+					strcpy(val, path);
+				} else {
+					/* other file like 'Makefile' */
+					val[0] = ' ';
+					strcpy(&val[1], path);
+				}
 				return val;
 			}
 			if (type == 'd') {
@@ -347,8 +488,11 @@ int	*length;
 	}
 	return NULL;
 }
+/*
+ * find_close: close iterator.
+ */
 void
-ffindclose(void)
+find_close(void)
 {
 	assert(opened == 1);
 	for (curp = &stack[0]; curp < topp; curp++)
@@ -365,173 +509,70 @@ ffindclose(void)
 /*----------------------------------------------------------------------*/
 static FILE	*ip;
 
+/*
+ * find_open: start iterator without GPATH.
+ */
 void
-ffindopen()
+find_open()
 {
-	char	*findcom, *p, *q;
-	STRBUF	*sb;
-	char	*sufflist = NULL;
-	char	*skiplist = NULL;
-	int	flags = REG_EXTENDED|REG_NEWLINE;
+	char	*findcom = "find . -type f -print";
 
 	assert(opened == 0);
 	opened = 1;
 
+	if (debug)
+		fprintf(stderr, "find com: %s\n", findcom);
 	/*
-	 * load icase_path option.
+	 * prepare regular expressions.
 	 */
-	if (getconfb("icase_path"))
-		flags |= REG_ICASE;
-#ifdef _WIN32
-	flags |= REG_ICASE;
-#endif /* _WIN32 */
-	sb = strbuf_open(0);
-	if (!getconfs("suffixes", sb))
-		die("cannot get suffixes data.");
-	sufflist = strdup(strbuf_value(sb));
-	if (!sufflist)
-		die("short of memory.");
-	trim(sufflist);
-	strbuf_reset(sb);
-	if (getconfs("skip", sb)) {
-		skiplist = strdup(strbuf_value(sb));
-		if (!skiplist)
-			die("short of memory.");
-		trim(skiplist);
-	}
-
-	strbuf_reset(sb);
-	strbuf_puts(sb, "find . \\( -type f -o -type l \\) \\(");
-	for (p = sufflist; p; ) {
-		char	*suff = p;
-		if ((p = locatestring(p, ",", MATCH_FIRST)) != NULL)
-			*p++ = 0;
-		strbuf_puts(sb, " -name '*.");
-		strbuf_puts(sb, suff);
-		strbuf_puts(sb, "'");
-		if (p)
-			strbuf_puts(sb, " -o");
-	}
-	strbuf_puts(sb, " \\) -print");
-	findcom = strbuf_value(sb);
-
-	if (skiplist) {
-		char	*reg;
-		STRBUF	*sbb = strbuf_open(0);
-		/*
-		 * construct regular expression.
-		 */
-		strbuf_putc(sbb, '(');	/* ) */
-		for (p = skiplist; p; ) {
-			char    *skipf = p;
-			if ((p = locatestring(p, ",", MATCH_FIRST)) != NULL)
-				*p++ = 0;
-			strbuf_putc(sbb, '/');
-			for (q = skipf; *q; q++) {
-				if (*q == '.')
-					strbuf_putc(sbb, '\\');
-				strbuf_putc(sbb, *q);
-			}
-			if (*(q - 1) != '/')
-				strbuf_putc(sbb, '$');
-			if (p)
-				strbuf_putc(sbb, '|');
-		}
-		strbuf_putc(sbb, ')');
-		reg = strbuf_value(sbb);
-		/*
-		 * compile regular expression.
-		 */
-		retval = regcomp(skip, reg, flags);
-		if (retval != 0)
-			die("cannot compile regular expression.");
-		strbuf_close(sbb);
-	} else {
-		skip = (regex_t *)0;
-	}
+	prepare_source();
+	prepare_skip();
 	if (!(ip = popen(findcom, "r")))
 		die("cannot execute find.");
-	strbuf_close(sb);
-	if (sufflist)
-		free(sufflist);
-	if (skiplist)
-		free(skiplist);
 }
+/*
+ * find_read: read path without GPATH.
+ *
+ *	r)		path
+ */
 char	*
-ffindread(length)
-int	*length;
+find_read(void)
 {
-	static char	path[MAXPATHLEN+1];
+	static char	val[MAXPATHLEN+2];
+	char	*path = &val[1];
 	char	*p;
 
 	assert(opened == 1);
 	while (fgets(path, MAXPATHLEN, ip)) {
-		if (!skip || regexec(skip, path, 0, 0, 0) != 0) {
-			/*
-			 * chop(path)
-			 */
-			p = path + strlen(path) - 1;
-			if (*p != '\n')
-				die("output of find(1) is wrong (findread).");
-			*p = 0;
-			if (length)
-				*length = p - path;
-			return path;
+		/*
+		 * chop(path)
+		 */
+		p = path + strlen(path) - 1;
+		if (*p != '\n')
+			die("output of find(1) is wrong (findread).");
+		*p = 0;
+		if (skipthisfile(path))
+			continue;
+		if (regexec(suff, path, 0, 0, 0) != 0) {
+			/* other file like 'Makefile' */
+			val[0] = ' ';
+			path = val;
 		}
+		return path;
 	}
 	return NULL;
 }
+/*
+ * find_close: close iterator.
+ */
 void
-ffindclose(void)
+find_close(void)
 {
 	assert(opened == 1);
 	pclose(ip);
 	opened = 0;
+	regfree(suff);
+	if (skip)
+		regfree(skip);
 }
 #endif /* !HAVE_FIND */
-
-/*
- * Faster findXXX() using GPATH.
- *
- * gfindxxx() does almost same with findxxx() but much faster,
- * because gfindxxx() use GPATH (file index).
- * If GPATH exist then you should use this.
- */
-#include "dbop.h"
-#include "gtagsop.h"
-
-static DBOP	*dbop;
-static int      opened;
-static int      first;
-static char	prefix[MAXPATHLEN+1];
-
-void
-gfindopen(dbpath, local)
-char	*dbpath;
-char	*local;
-{
-	assert(opened == 0);
-	assert(first == 0);
-	dbop = dbop_open(makepath(dbpath, dbname(GPATH), NULL), 0, 0, 0);
-	if (dbop == NULL)
-		die("GPATH not found.");
-	strcpy(prefix, (local) ? local : "./");
-	opened = 1;
-	first = 1;
-}
-char	*
-gfindread()
-{
-	assert(opened == 1);
-	if (first) {
-		first = 0;
-		return dbop_first(dbop, prefix, NULL, DBOP_KEY | DBOP_PREFIX);
-	}
-	return dbop_next(dbop);
-}
-void
-gfindclose(void)
-{
-	dbop_close(dbop);
-	opened = first = 0;
-}
