@@ -1,7 +1,7 @@
 /*
  * Copyright (c) 1996, 1997, 1998, 1999
  *             Shigio Yamaguchi. All rights reserved.
- * Copyright (c) 1999, 2000
+ * Copyright (c) 1999, 2000, 2002
  *             Tama Communications Corporation. All rights reserved.
  *
  * This file is part of GNU GLOBAL.
@@ -42,7 +42,7 @@
 #include "strbuf.h"
 #include "usable.h"
 
-#ifdef _WIN32
+#if defined(_WIN32) || defined(__DJGPP__)
 static const char *suffix[] = {".exe", ".com", ".bat",};
 #endif
 
@@ -60,7 +60,8 @@ char	*command;
 	STRBUF	*sb;
 	char *p, *dir;
 	static char path[MAXPATHLEN+1];
-#ifdef _WIN32
+
+#if defined(_WIN32) || defined(__DJGPP__)
 	int	i, lim = sizeof(suffix)/sizeof(char *);
 #endif
 
@@ -73,7 +74,15 @@ char	*command;
 		return NULL;
 	}
 	/*
-	 * Locate the command for each path in GTAGSLIBPATH.
+	 * If found in BINDIR then use it.
+	 */
+	if (test("fx", makepath(BINDIR, command, NULL))) {
+		strncpy(path, makepath(BINDIR, command, NULL), sizeof(path));
+		path[sizeof(path) - 1] = '\0';
+		return path;
+	}
+	/*
+	 * Locate the command for each path in PATH.
 	 */
 	*path = 0;
 	/* Don't use fixed length buffer for environment variable
@@ -86,13 +95,15 @@ char	*command;
 		if ((p = locatestring(p, PATHSEP, MATCH_FIRST)) != NULL)
 			*p++ = 0;
 		if (test("fx", makepath(dir, command, NULL))) {
-			strcpy(path, makepath(dir, command, NULL));
+			strncpy(path, makepath(dir, command, NULL), sizeof(path));
+			path[sizeof(path) - 1] = '\0';
 			goto finish;
 		}
-#if defined(_WIN32)
+#if defined(_WIN32) || defined(__DJGPP__)
 		for (i = 0; i < lim; i++)
 			if (test("f", makepath(dir, command, suffix[i]))) {
-				strcpy(path, makepath(dir, command, suffix[i]));
+				strncpy(path, makepath(dir, command, suffix[i]), sizeof(path));
+				path[sizeof(path) - 1] = '\0';
 				goto finish;
 			}
 #endif
